@@ -3,14 +3,14 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2005 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2009 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  * 
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  * 
  * GAWK is distributed in the hope that it will be useful,
@@ -39,20 +39,20 @@ is_blank(int c)
 typedef void (* Setfunc) P((long, char *, long, NODE *));
 
 static long (*parse_field) P((long, char **, int, NODE *,
-			     Regexp *, Setfunc, NODE *));
+			     Regexp *, Setfunc, NODE *, int));
 static void rebuild_record P((void));
 static long re_parse_field P((long, char **, int, NODE *,
-			     Regexp *, Setfunc, NODE *));
+			     Regexp *, Setfunc, NODE *, int));
 static long def_parse_field P((long, char **, int, NODE *,
-			      Regexp *, Setfunc, NODE *));
+			      Regexp *, Setfunc, NODE *, int));
 static long posix_def_parse_field P((long, char **, int, NODE *,
-			      Regexp *, Setfunc, NODE *));
+			      Regexp *, Setfunc, NODE *, int));
 static long null_parse_field P((long, char **, int, NODE *,
-			     Regexp *, Setfunc, NODE *));
+			     Regexp *, Setfunc, NODE *, int));
 static long sc_parse_field P((long, char **, int, NODE *,
-			     Regexp *, Setfunc, NODE *));
+			     Regexp *, Setfunc, NODE *, int));
 static long fw_parse_field P((long, char **, int, NODE *,
-			     Regexp *, Setfunc, NODE *));
+			     Regexp *, Setfunc, NODE *, int));
 static void set_element P((long num, char * str, long len, NODE *arr));
 static void grow_fields_arr P((long num));
 static void set_field P((long num, char *str, long len, NODE *dummy));
@@ -166,6 +166,7 @@ rebuild_record()
 	cops = ops;
 	ops[0] = '\0';
 	for (i = 1;  i <= NF; i++) {
+		free_wstr(fields_arr[i]);
 		tmp = fields_arr[i];
 		/* copy field */
 		if (tmp->stlen == 1)
@@ -212,6 +213,9 @@ rebuild_record()
 			n->stptr = cops;
 			unref(fields_arr[i]);
 			fields_arr[i] = n;
+#ifdef MBS_SUPPORT
+			assert((n->flags & WSTRCUR) == 0);
+#endif
 		}
 		cops += fields_arr[i]->stlen + ofslen;
 	}
@@ -360,12 +364,14 @@ re_parse_field(long up_to,	/* parse only up to this field number */
 	NODE *fs ATTRIBUTE_UNUSED,
 	Regexp *rp,
 	Setfunc set,	/* routine to set the value of the parsed field */
-	NODE *n)
+	NODE *n,
+	int in_middle)
 {
 	register char *scan = *buf;
 	register long nf = parse_high_water;
 	register char *field;
 	register char *end = scan + len;
+	int regex_flags = RE_NEED_START;
 #ifdef MBS_SUPPORT
 	size_t mbclen = 0;
 	mbstate_t mbs;
@@ -373,6 +379,9 @@ re_parse_field(long up_to,	/* parse only up to this field number */
 		memset(&mbs, 0, sizeof(mbstate_t));
 #endif
 
+	if (in_middle)
+		regex_flags |= RE_NO_BOL;
+
 	if (up_to == UNLIMITED)
 		nf = 0;
 	if (len == 0)
@@ -383,8 +392,9 @@ re_parse_field(long up_to,	/* parse only up to this field number */
 			scan++;
 	field = scan;
 	while (scan < end
-	       && research(rp, scan, 0, (end - scan), RE_NEED_START) != -1
+	       && research(rp, scan, 0, (end - scan), regex_flags) != -1
 	       && nf < up_to) {
+		regex_flags |= RE_NO_BOL;
 		if (REEND(rp, scan) == RESTART(rp, scan)) {   /* null match */
 #ifdef MBS_SUPPORT
 			if (gawk_mb_cur_max > 1)	{
@@ -435,7 +445,8 @@ def_parse_field(long up_to,	/* parse only up to this field number */
 	NODE *fs,
 	Regexp *rp ATTRIBUTE_UNUSED,
 	Setfunc set,	/* routine to set the value of the parsed field */
-	NODE *n)
+	NODE *n,
+	int in_middle ATTRIBUTE_UNUSED)
 {
 	register char *scan = *buf;
 	register long nf = parse_high_water;
@@ -502,7 +513,8 @@ posix_def_parse_field(long up_to,	/* parse only up to this field number */
 	NODE *fs,
 	Regexp *rp ATTRIBUTE_UNUSED,
 	Setfunc set,	/* routine to set the value of the parsed field */
-	NODE *n)
+	NODE *n,
+	int in_middle ATTRIBUTE_UNUSED)
 {
 	register char *scan = *buf;
 	register long nf = parse_high_water;
@@ -566,7 +578,8 @@ null_parse_field(long up_to,	/* parse only up to this field number */
 	NODE *fs ATTRIBUTE_UNUSED,
 	Regexp *rp ATTRIBUTE_UNUSED,
 	Setfunc set,	/* routine to set the value of the parsed field */
-	NODE *n)
+	NODE *n,
+	int in_middle ATTRIBUTE_UNUSED)
 {
 	register char *scan = *buf;
 	register long nf = parse_high_water;
@@ -614,7 +627,8 @@ sc_parse_field(long up_to,	/* parse only up to this field number */
 	NODE *fs,
 	Regexp *rp ATTRIBUTE_UNUSED,
 	Setfunc set,	/* routine to set the value of the parsed field */
-	NODE *n)
+	NODE *n,
+	int in_middle ATTRIBUTE_UNUSED)
 {
 	register char *scan = *buf;
 	register char fschar;
@@ -691,21 +705,60 @@ fw_parse_field(long up_to,	/* parse only up to this field number */
 	NODE *fs ATTRIBUTE_UNUSED,
 	Regexp *rp ATTRIBUTE_UNUSED,
 	Setfunc set,	/* routine to set the value of the parsed field */
-	NODE *n)
+	NODE *n,
+	int in_middle ATTRIBUTE_UNUSED)
 {
 	register char *scan = *buf;
 	register long nf = parse_high_water;
 	register char *end = scan + len;
+#ifdef MBS_SUPPORT
+	int nmbc;
+	size_t mbclen;
+	size_t mbslen;
+	size_t lenrest;
+	char *mbscan;
+	mbstate_t mbs;
+
+	memset(&mbs, 0, sizeof(mbstate_t));
+#endif
 
 	if (up_to == UNLIMITED)
 		nf = 0;
 	if (len == 0)
 		return nf;
 	for (; nf < up_to && (len = FIELDWIDTHS[nf+1]) != -1; ) {
-		if (len > end - scan)
-			len = end - scan;
-		(*set)(++nf, scan, (long) len, n);
-		scan += len;
+#ifdef MBS_SUPPORT
+		if (gawk_mb_cur_max > 1) {
+			nmbc = 0;
+			mbslen = 0;
+			mbscan = scan;
+			lenrest = end - scan;
+			while (nmbc < len && mbslen < lenrest) {
+				mbclen = mbrlen(mbscan, end - mbscan, &mbs);
+				if (   mbclen == 1
+				    || mbclen == (size_t) -1
+				    || mbclen == (size_t) -2
+				    || mbclen == 0) {
+					/* We treat it as a singlebyte character.  */
+		    			mbclen = 1;
+				}
+				if (mbclen <= end - mbscan) {
+					mbscan += mbclen;
+		    			mbslen += mbclen;
+		    			++nmbc;
+				}
+	    		}
+			(*set)(++nf, scan, (long) mbslen, n);
+			scan += mbslen;
+		}
+		else
+#endif
+		{
+			if (len > end - scan)
+				len = end - scan;
+			(*set)(++nf, scan, (long) len, n);
+			scan += len;
+		}
 	}
 	if (len == -1)
 		*buf = end;
@@ -721,6 +774,7 @@ fw_parse_field(long up_to,	/* parse only up to this field number */
 NODE **
 get_field(register long requested, Func_ptr *assign)
 {
+	int in_middle = FALSE;
 	/*
 	 * if requesting whole line but some other field has been altered,
 	 * then the whole line must be rebuilt
@@ -733,7 +787,7 @@ get_field(register long requested, Func_ptr *assign)
 		    			fields_arr[0]->stlen -
 					(parse_extent - fields_arr[0]->stptr),
 		    			save_FS, FS_regexp, set_field,
-					(NODE *) NULL);
+					(NODE *) NULL, in_middle);
 				parse_high_water = NF;
 			}
 			rebuild_record();
@@ -758,9 +812,11 @@ get_field(register long requested, Func_ptr *assign)
 		 */
 		if (parse_high_water == 0)	/* starting at the beginning */
 			parse_extent = fields_arr[0]->stptr;
+		else
+			in_middle = TRUE;
 		parse_high_water = (*parse_field)(requested, &parse_extent,
 		     fields_arr[0]->stlen - (parse_extent - fields_arr[0]->stptr),
-		     save_FS, FS_regexp, set_field, (NODE *) NULL);
+		     save_FS, FS_regexp, set_field, (NODE *) NULL, in_middle);
 
 		/*
 		 * if we reached the end of the record, set NF to the number of
@@ -809,7 +865,7 @@ do_split(NODE *tree)
 	NODE *src, *arr, *sep, *fs, *src2, *fs2, *tmp;
 	char *s;
 	long (*parseit) P((long, char **, int, NODE *,
-			 Regexp *, Setfunc, NODE *));
+			 Regexp *, Setfunc, NODE *, int));
 	Regexp *rp = NULL;
 
 	src = force_string(tree_eval(tree->lnode));
@@ -828,7 +884,7 @@ do_split(NODE *tree)
 		/*
 		 * Evaluate sep if it may have side effects.
 		 */
-		if ((sep->re_flags & (FS_DFLT|CONST)) == 0)
+		if ((sep->re_flags & (FS_DFLT|CONSTANT)) == 0)
 			free_temp(tree_eval(sep->re_exp));
 		/*
 		 * And now we can safely turn off the array.
@@ -852,7 +908,7 @@ do_split(NODE *tree)
 				warned = TRUE;
 				lintwarn(_("split: null string for third arg is a gawk extension"));
 			}
-		} else if (fs->stlen == 1 && (sep->re_flags & CONST) == 0) {
+		} else if (fs->stlen == 1 && (sep->re_flags & CONSTANT) == 0) {
 			if (fs->stptr[0] == ' ') {
 				if (do_posix)
 					parseit = posix_def_parse_field;
@@ -882,7 +938,7 @@ do_split(NODE *tree)
 
 	s = src2->stptr;
 	tmp = tmp_number((AWKNUM) (*parseit)(UNLIMITED, &s, (int) src2->stlen,
-					     fs2, rp, set_element, arr));
+					     fs2, rp, set_element, arr, FALSE));
 	unref(src2);
 	unref(fs2);
 	return tmp;
@@ -897,7 +953,7 @@ set_FIELDWIDTHS()
 	char *end;
 	register int i;
 	static int fw_alloc = 4;
-	static int warned = FALSE;
+	static short warned = FALSE;
 	extern unsigned long strtoul P((const char *, char **endptr, int base));
 
 	if (do_lint && ! warned) {
@@ -916,13 +972,13 @@ set_FIELDWIDTHS()
 
 	parse_field = fw_parse_field;
 	scan = force_string(FIELDWIDTHS_node->var_value)->stptr;
-	end = scan + 1;
+
 	if (FIELDWIDTHS == NULL)
 		emalloc(FIELDWIDTHS, int *, fw_alloc * sizeof(int), "set_FIELDWIDTHS");
 	FIELDWIDTHS[0] = 0;
 	for (i = 1; ; i++) {
 		unsigned long int tmp;
-		if (i >= fw_alloc) {
+		if (i + 1 >= fw_alloc) {
 			fw_alloc *= 2;
 			erealloc(FIELDWIDTHS, int *, fw_alloc * sizeof(int), "set_FIELDWIDTHS");
 		}
@@ -935,13 +991,16 @@ set_FIELDWIDTHS()
 			fatal(_("invalid FIELDWIDTHS value, near `%s'"),
 			      scan);
 
+		if (*scan == '\0')
+			break;
+
 		/* Detect an invalid base-10 integer, a valid value that
 		   is followed by something other than a blank or '\0',
 		   or a value that is not in the range [1..INT_MAX].  */
 		errno = 0;
 		tmp = strtoul(scan, &end, 10);
 		if (errno != 0
-		    || !(*end == '\0' || is_blank(*end))
+		    || (*end != '\0' && ! is_blank(*end))
 		    || !(0 < tmp && tmp <= INT_MAX))
 			fatal(_("invalid FIELDWIDTHS value, near `%s'"),
 			      scan);
@@ -954,7 +1013,9 @@ set_FIELDWIDTHS()
 		if (*scan == '\0')
 			break;
 	}
-	FIELDWIDTHS[i] = -1;
+	if (i == 1)	/* empty string! */
+		i--;
+	FIELDWIDTHS[i+1] = -1;
 
 	update_PROCINFO("FS", "FIELDWIDTHS");
 }
@@ -1027,6 +1088,8 @@ choose_fs_function:
 			lintwarn(_("null string for `FS' is a gawk extension"));
 		}
 	} else if (fs->stlen > 1) {
+		if (do_lint_old)
+			warning(_("old awk does not support regexps as value of `FS'"));
 		parse_field = re_parse_field;
 	} else if (RS_is_null) {
 		/* we know that fs->stlen <= 1 */
